# -*- coding: utf-8 -*-

import functools
import json

# pylint: disable=unused-import
from .poistuvat import (
  csrf_tarkistus,
  json_viestiliikenne,
  protokolla,
)
# pylint: enable=unused-import


class Koriste:

  __wrapped__: callable

  def __new__(cls, f=None, **kwargs):
    '''
    Kääritään olio annetun `f`-toteutuksen mukaan.
    Mikäli tätä ei annettu, palautetaan olion sijaan sulkeuma.
    '''
    # pylint: disable=invalid-name
    if f is None:
      return functools.partial(cls, **kwargs)
    return functools.wraps(
      f
    )(super().__new__(cls))
    # def __new__

  def __init__(self, f):
    '''
    Parametriä `f` käytetään vain `__new__`-metodissa.
    '''
    # pylint: disable=invalid-name, unused-argument
    super().__init__()
    # def __init__

  def __call__(self, request, *args, **kwargs):
    return self.__wrapped__(request, *args, **kwargs)
    # def __call__

  # class Koriste


class OriginPoikkeus(Koriste):
  ''' Ohita Origin-otsakkeen tarkistus Websocket-pyynnön yhteydessä. '''
  origin_poikkeus = True
  # def origin_poikkeus


class JsonLiikenne(Koriste):

  def __init__(
    self,
    websocket, *,
    loads=None,
    dumps=None,
  ):
    super().__init__(websocket)
    self.loads = loads or {}
    self.dumps = dumps or {}
    # def __init__

  async def __call__(self, request, *args, **kwargs):
    @functools.wraps(request.receive)
    async def receive():
      return json.loads(
        await receive.__wrapped__(),
        **self.loads
      )
    @functools.wraps(request.send)
    async def send(s):
      return await send.__wrapped__(
        json.dumps(s, **self.dumps)
      )
    request.receive = receive
    request.send = send
    try:
      return await self.__wrapped__(
        request, *args, **kwargs
      )
    finally:
      request.receive = receive.__wrapped__
      request.send = send.__wrapped__
    # async def __call__

  # class JsonLiikenne


class CsrfKattely(Koriste):

  def __init__(
    self,
    websocket, *,
    csrf_avain=None,
    virhe_avain=None
  ):
    super().__init__(websocket)
    self.csrf_avain = csrf_avain
    self.virhe_avain = virhe_avain
    # def __init__

  async def __call__(self, request, *args, **kwargs):
    try:
      kattely = await request.receive()
    except ValueError:
      virhe = 'Yhteyden muodostus epäonnistui!'
      return await request.send({
        self.virhe_avain: virhe
      } if self.virhe_avain is not None else virhe)
    if getattr(request, '_dont_enforce_csrf_checks', False):
      # Ohitetaan CSRF-tarkistus testauksen yhteydessä.
      pass
    elif not request.tarkista_csrf(
      kattely.get(self.csrf_avain)
      if self.csrf_avain else kattely
    ):
      virhe = 'CSRF-avain puuttuu tai se on virheellinen!'
      return await request.send({
        self.virhe_avain: virhe
      } if self.virhe_avain else virhe)

    request._csrf_kattely = kattely
    return await self.__wrapped__(
      request, *args, **kwargs
    )
    # async def __call__

  # class CsrfKattely



# -*- coding: utf-8 -*-

import asyncio
from contextlib import asynccontextmanager
import logging

from asgiref.sync import sync_to_async, iscoroutinefunction

import django
from django.conf import settings
from django.core.handlers.asgi import ASGIHandler
from django.core import signals
from django.test.utils import override_settings
from django.urls import set_script_prefix, set_urlconf

from pistoke.protokolla import _WebsocketProtokolla
from pistoke.pyynto import WebsocketPyynto


loki = logging.getLogger('django.' + __name__)


class WebsocketVirhe(RuntimeError):
  ''' Virheellinen konteksti Websocket-pyynnön käsittelyssä (WSGI). '''


class WebsocketKasittelija(ASGIHandler):
  '''
  Saapuvien Websocket-pyyntöjen (istuntojen) käsittelyrutiini.
  '''

  nosta_syotetta_ei_luettu: bool = False

  def __new__(cls, *args, **kwargs):
    '''
    Alusta Django ennen käsittelyrutiinin luontia.

    Vrt. get_asgi_application().
    '''
    django.setup(set_prefix=False)
    return super().__new__(cls)
    # def __new__

  @asynccontextmanager
  async def _django_pyynto(self, scope):
    # Tehdään Django-rutiinitoimet per saapuva pyyntö.
    await sync_to_async(
      signals.request_started.send,
      thread_sensitive=True
    )(
      sender=self.__class__, scope=scope
    )
    try:
      yield
    finally:
      try:
        await asyncio.shield(sync_to_async(
          signals.request_finished.send,
          thread_sensitive=True
        )(
          sender=self.__class__
        ))
      except asyncio.CancelledError:
        pass
    # def _django_pyynto

  async def __call__(self, scope, receive, send):
    '''
    Asynkroninen, pyyntökohtainen kutsu.

    Vrt. django.core.handlers.asgi:ASGIHandler.__call__
    '''
    assert scope['type'] == 'websocket'

    if hasattr(self, 'get_script_prefix'):  # Django 4.2-
      set_script_prefix(self.get_script_prefix(scope))
    else:  # Django 5.0+
      from django.core.handlers.asgi import get_script_prefix
      set_script_prefix(get_script_prefix(scope))

    # Tehdään Django-rutiinitoimet per saapuva pyyntö.
    async with self._django_pyynto(scope):
      # Muodostetaan WS-pyyntöolio.
      request = WebsocketPyynto(scope, receive, send)

      # Hae käsittelevä näkymärutiini tai mahdollinen virheviesti.
      # Tämä kutsuu mahdollisten avaavien välikkeiden (middleware) ketjua
      # ja lopuksi alla määriteltyä `_get_response_async`-metodia.
      # Metodi suorittaa ensin Websocket-kättelyn loppuun ja sen jälkeen
      # URL-taulun mukaisen näkymäfunktion (async def websocket(...): ...).
      nakyma = await self.get_response_async(request)

      if asyncio.iscoroutine(nakyma):
        await nakyma

      else:
        # Ota yhteyspyyntö vastaan ja evää se.
        # Tällöin asiakaspäähän palautuu HTTP 403 Forbidden.
        avaus = await request.receive()
        assert avaus.get('type') == 'websocket.connect'
        await request.send({'type': 'websocket.close'})
        # if not asyncio.iscoroutine
      # async with self._django_pyynto
    # async def __call__

  def load_middleware(self, is_async=False):
    '''
    Ajetaan vain muunnostaulun mukaiset Websocket-pyynnölle käyttöön
    otettavat ohjaimet.
    '''
    from pistoke.ohjain import websocket_ohjaimet
    with override_settings(MIDDLEWARE=websocket_ohjaimet):
      super().load_middleware(is_async=is_async)
    # def load_middleware

  # Synkroniset pyynnöt nostavat poikkeuksen.
  def get_response(self, request):
    raise WebsocketVirhe
  def _get_response(self, request):
    raise WebsocketVirhe

  async def get_response_async(self, request):
    ''' Ohitetaan paluusanoman käsittelyyn liittyvät funktiokutsut. '''
    set_urlconf(settings.ROOT_URLCONF)
    return await self._middleware_chain(request)
    # async def get_response_async

  async def _get_response_async(self, request):
    ''' Ohitetaan paluusanoman käsittelyyn liittyvät funktiokutsut. '''
    # pylint: disable=not-callable, protected-access
    from pistoke.protokolla import WebsocketProtokolla
    @WebsocketProtokolla
    async def evatty(*args, **kwargs): pass

    callback, callback_args, callback_kwargs = self.resolve_request(request)
    for middleware_method in self._view_middleware:
      vastaus = await middleware_method(
        request, callback, callback_args, callback_kwargs
      )
      if vastaus is not None:
        loki.debug(
          'Ohjainketju palautti HTTP-vastauksen %r.',
          vastaus,
        )
        return evatty
      # for middleware_method in self._view_middleware

    # Mikäli `callback` on asynkroninen funktio (tai kääre),
    # palautetaan sen tuottama alirutiini.
    if iscoroutinefunction(callback) \
    or iscoroutinefunction(getattr(callback, '__call__', callback)):
      return callback(
        request, *callback_args, **callback_kwargs
      )

    # Mikäli `callback` on synkroninen funktio (View.dispatch tai vastaava),
    # kutsutaan sitä pääsäikeessä.
    if callable(callback):
      nakyma = await sync_to_async(
        callback,
        thread_sensitive=True
      )(
        request, *callback_args, **callback_kwargs
      )
      # Mikäli tuloksena palautuu `async def websocket(...)`-metodin
      # (tai vastaavan) tuottama alirutiini, palautetaan se.
      if asyncio.iscoroutine(nakyma):
        async def _nakyma():
          try:
            return await nakyma
          except _WebsocketProtokolla.SyotettaEiLuettu:
            if self.nosta_syotetta_ei_luettu:
              raise
        return _nakyma()
      # Muussa tapauksessa kyse voi olla esim. uudelleenohjauksesta
      # kirjautumissivulle.
      # Evätään tällöin Websocket-pyyntö.
      loki.debug(
        'Websocket-näkymä %r palautti alirutiinin sijaan arvon %r.',
        getattr(callback, 'view_class', callback),
        nakyma,
      )
      return evatty
    else:
      raise ValueError(
        f'Websocket-näkymä {callback!r} ei ole kelvollinen funktio.'
      )
    # async def _get_response_async

  # class WebsocketKasittelija



# -*- coding: utf-8 -*-

from django.conf import settings
from django.core.handlers.asgi import ASGIRequest
from django.http import QueryDict
from django import VERSION as django_version


class WebsocketPyynto(ASGIRequest):
  '''
  Yksittäisen Websocket-pyynnön (istunnon) tiedot.

  Huomaa, että __init__ ei kutsu super-toteutusta.
  '''
  # pylint: disable=too-many-instance-attributes
  # pylint: disable=method-hidden
  # pylint: disable=invalid-name

  POST = QueryDict()
  FILES = {}

  def __init__(self, scope, receive, send):
    # pylint: disable=super-init-not-called
    self.scope = scope
    self.receive = receive
    self.send = send
    self._post_parse_error = False
    self._read_started = False
    self.resolver_match = None
    self.script_name = self.scope.get('root_path', '')
    if self.script_name and scope['path'].startswith(self.script_name):
      self.path_info = scope['path'][len(self.script_name):]
    else:
      self.path_info = scope['path']
    if self.script_name:
      self.path = '%s/%s' % (
        self.script_name.rstrip('/'),
        self.path_info.replace('/', '', 1),
      )
    else:
      self.path = scope['path']

    self.method = 'Websocket'

    query_string = self.scope.get('query_string', '')
    if isinstance(query_string, bytes):
      query_string = query_string.decode()
    self.META = {
      'REQUEST_METHOD': self.method,
      'QUERY_STRING': query_string,
      'SCRIPT_NAME': self.script_name,
      'PATH_INFO': self.path_info,
      'wsgi.multithread': True,
      'wsgi.multiprocess': True,
    }
    if self.scope.get('client'):
      self.META['REMOTE_ADDR'] = self.scope['client'][0]
      self.META['REMOTE_HOST'] = self.META['REMOTE_ADDR']
      self.META['REMOTE_PORT'] = self.scope['client'][1]
    if self.scope.get('server'):
      self.META['SERVER_NAME'] = self.scope['server'][0]
      self.META['SERVER_PORT'] = str(self.scope['server'][1])
    else:
      self.META['SERVER_NAME'] = 'unknown'
      self.META['SERVER_PORT'] = '0'
    for name, value in self.scope.get('headers', []):
      name = name.decode('latin1')
      corrected_name = 'HTTP_%s' % name.upper().replace('-', '_')
      value = value.decode('latin1')
      if corrected_name in self.META:
        value = self.META[corrected_name] + ',' + value
      self.META[corrected_name] = value

    self.resolver_match = None
    # def __init__

  # Lisätään `tarkista_csrf`-metodin toteutus Django-versiokohtaisesti.
  # pylint: disable=no-name-in-module, undefined-variable
  if django_version >= (4, 1):
    def tarkista_csrf(self, csrf_token):
      '''
      Vrt. django.middleware.csrf:CsrfMiddleware._check_token.

      Huomaa, että tässä käytetään `self.META["CSRF_COOKIE"]`-avainta,
      jonka `CsrfMiddleware.process_request` asettaa.
      '''
      # pylint: disable=no-member
      from django.middleware.csrf import (
        _check_token_format,
        _does_token_match,
        InvalidTokenFormat,
      )
      if csrf_token is None:
        return False
      if (csrf_secret := self.META.get('CSRF_COOKIE')) is None:
        return False
      try:
        _check_token_format(csrf_token)
        _check_token_format(csrf_secret)
      except InvalidTokenFormat:
        return False
      return _does_token_match(csrf_token, csrf_secret)
      # def tarkista_csrf
    # if django_version >= 4.1
  else:
    def tarkista_csrf(self, csrf_token):
      if django_version >= (4, 0):
        from django.middleware.csrf import _does_token_match
      else:
        from django.middleware.csrf import (
          _compare_masked_tokens as _does_token_match
        )
      from django.middleware.csrf import _sanitize_token
      return csrf_token \
      and self.META.get('CSRF_COOKIE') \
      and _does_token_match(
        _sanitize_token(csrf_token),
        self.META.get('CSRF_COOKIE'),
      )
    # else (if django_version < 4.1)
  # pylint: enable=no-name-in-module, undefined-variable
  tarkista_csrf.__doc__ = '''
  Tarkista pyyntödatan mukana saatu CSRF-tunniste evästeenä
  annettua tunnistetta vasten.
  '''

  async def __aiter__(self):
    while True:
      yield await self.receive()
    # async def __aiter__

  # class WebsocketPyynto



# -*- coding: utf-8 -*-

from django.apps import AppConfig

from .nakyma import (
  WebsocketNakyma,
)

try:
  from .protokolla import (
    WebsocketProtokolla,
    WebsocketAliprotokolla,
  )
except ImportError:
  pass


class Pistoke(AppConfig):
  name = 'pistoke'



# -*- coding: utf-8 -*-

'''Django-välikkeet (Middleware) Websocket-ympäristössä.

HTTP-pyynnöllä ajettava `WebsocketOhjain` asettaa pyynnölle `websocket`-
määreen silloin, kun Websocket-yhteys on käytettävissä, so. pyyntö on
tullut sisään Djangon ASGI-käsittelijän kautta. Silloin on voimassa:
- kun pyydetty osoite on http(s)://palvelimen.osoite,
- määre `request.websocket` on ws(s)://palvelimen.osoite.

Lisäksi tämä moduuli määrittelee (`WEBSOCKET_MIDDLEWARE`) luettelon,
Django-projektiasetuksia mukaillen, välikkeistä jotka voidaan ajaa
saapuville Websocket-pyynnöille.

Käytännössä nämä ovat projektiasetusten mukaiset HTTP-välikkeet ilman
paluusanoman käsittelytoteutusta (`process_response`).

Lisäksi seuraavien välikkeiden toimintaa on mukautettu:
- CsrfMiddleware: ohitetaan POST-datan (jota ei ole) automaattinen,
  oletusarvoinen tarkistus; lisätään pyyntökohtainen
  `request.tarkista_csrf`-metodi.
- OriginTarkistus: ajetaan vain Websocket-pyynnöille, tarkistetaan
  täsmääkö HTTP-`Origin`-otsake Django-`ALLOWED_HOSTS`-luetteloon.
  Tämä tarkistus ohitetaan, mikäli näkymäfunktiolle on asetettu määre
  `origin_poikkeus` (ks. `pistoke.tyokalut.origin_poikkeus`).
'''

import logging
from urllib.parse import urlparse

from asgiref.sync import iscoroutinefunction, markcoroutinefunction

from django.conf import settings
from django.contrib.sessions.middleware import SessionMiddleware
from django.core.handlers.asgi import ASGIRequest
from django.http.request import split_domain_port, validate_host
from django.http import HttpResponseForbidden
from django.middleware.csrf import CsrfViewMiddleware
from django.utils.decorators import sync_and_async_middleware
from django.utils.module_loading import import_string
from django.utils.log import log_response

import mmaare


logger = logging.getLogger('django.pistoke.origin')


class OhitaPaluusanoma:
  ''' Ohjain, joka ohittaa paluusanoman käsittelyn. '''

  def process_response(self, request, response):
    # pylint: disable=unused-argument
    return response

  # class OhitaPaluusanoma


class Ohjain:
  '''
  Ohjain, joka näyttää asynkroniselta ulospäin silloin, kun signaalitien
  seuraava ohjain on asynkroninen.
  '''

  def __init__(self, get_response):
    self.get_response = get_response
    if iscoroutinefunction(self.get_response):
      markcoroutinefunction(self)
    # def __init__

  # class Ohjain


@sync_and_async_middleware
class WebsocketOhjain(Ohjain):
  '''
  Ohjain, joka asettaa pyynnölle URI-määritteen `websocket` siten,
  että siihen tehdyt pyynnöt ohjataan ajossa olevaan ASGI-käsittelijään.

  Mikäli HTTP-yhteys on salattu (HTTPS),
  käytetään salattua Websocket-yhteyttä (WSS).

  Mikäli pyyntö ei ole ASGI-pohjainen, ei `websocket`-yhteys ole
  käytettävissä eikä em. määritettä aseteta.
  '''

  def __call__(self, request):
    # Huomaa, että funktiokutsu palauttaa joko synkronisen tuloksen
    # tai asynkronisen alirutiinin sen mukaan, mitä signaalitien
    # seuraava ohjain palauttaa.
    if isinstance(request, ASGIRequest):
      request.websocket = (
        f'{"wss" if request.is_secure() else "ws"}://{request.get_host()}'
      )
    return self.get_response(request)
    # def __call__

  # class WebsocketOhjain


@sync_and_async_middleware
class OriginVaatimus(Ohjain):
  '''
  Ohjain, joka tarkistaa mahdollisen Websocket-pyynnöllä annetun
  Origin-otsakkeen vastaavasti kuin Django-CSRF-ohjain tarkistaa
  sen POST-pyynnöllä.
  '''

  def __call__(self, request):
    return self.get_response(request)
    # def __call__

  def process_view(self, request, callback, callback_args, callback_kwargs):
    # pylint: disable=unused-argument
    assert request.method == 'Websocket'
    if getattr(callback, 'origin_poikkeus', False):
      return None
    elif 'HTTP_ORIGIN' not in request.META:
      return None
    origin = split_domain_port(
      urlparse(request.META['HTTP_ORIGIN']).netloc.lower()
    )[0]
    if not validate_host(origin, settings.ALLOWED_HOSTS):
      virhe = 'Websocket: Origin=%r ei vastaa ALLOWED_HOSTS-asetusta.' % origin
      response = HttpResponseForbidden(virhe)
      log_response(
        virhe,
        request=request,
        response=response,
        logger=logger,
      )
      return response
      # if not validate_host
    return None
    # def process_view

  # class OriginVaatimus


class CsrfOhjain(OhitaPaluusanoma, CsrfViewMiddleware):
  def process_view(self, request, callback, callback_args, callback_kwargs):
    '''
    Ohitetaan tavanomainen CSRF-tarkistus POST-datasta.
    '''
  # class CsrfOhjain


class IstuntoOhjain(OhitaPaluusanoma, SessionMiddleware):
  pass


# Muunnostaulu Websocket-pyyntöihin sovellettavista
# Middleware-ohjaimista.
# Muut kuin tässä mainitut ohjaimet periytetään automaattisesti
# `OhitaPaluusanoma`-saateluokasta.
WEBSOCKET_MIDDLEWARE = {
  # Ohitetaan kokonaan.
  'corsheaders.middleware.CorsMiddleware': False,
  'debug_toolbar.middleware.DebugToolbarMiddleware': False,
  'django.middleware.gzip.GZipMiddleware': False,
  'django.middleware.clickjacking.XFrameOptionsMiddleware': False,
  'django_hosts.middleware.HostsResponseMiddleware': False,
  'request_logging.middleware.LoggingMiddleware': False,

  # Korvataan muunnoksella.
  'django.middleware.csrf.CsrfViewMiddleware': 'pistoke.ohjain.CsrfOhjain',
  'django.contrib.sessions.middleware.SessionMiddleware': \
    'pistoke.ohjain.IstuntoOhjain',

  # Suoritetaan sellaisenaan.
  'pistoke.ohjain.WebsocketOhjain': True,
}


def sovita_ohjain_websocket_pyynnolle(ohjain):
  if isinstance(ohjain, str):
    ohjain = import_string(ohjain)
  if isinstance(ohjain, type):
    return type(
      ohjain.__name__,
      (OhitaPaluusanoma, ohjain),
      {}
    )
  elif callable(ohjain):
    return ohjain
  else:
    raise NotImplementedError
  # def sovita_ohjain_websocket_pyynnolle

@mmaare
def __websocket_ohjaimet(moduuli):
  def _websocket_ohjaimet():
    for ohjain in settings.MIDDLEWARE:
      muunnos = WEBSOCKET_MIDDLEWARE.get(ohjain, None)
      if muunnos is False:
        continue
      elif muunnos is True:
        yield ohjain
        continue
      elif muunnos is not None:
        yield muunnos
        continue
      ohjain = sovita_ohjain_websocket_pyynnolle(ohjain)
      setattr(moduuli, ohjain.__name__, ohjain)
      yield '.'.join((__name__, ohjain.__name__))
    yield 'pistoke.ohjain.OriginVaatimus'
  return list(_websocket_ohjaimet())
  # def __websocket_ohjaimet



# -*- coding: utf-8 -*-

for i, sovellus in enumerate(INSTALLED_APPS):
  if sovellus == 'django.contrib.staticfiles':
    INSTALLED_APPS.insert(i, 'pistoke.Pistoke')
    break
  elif sovellus == 'pistoke.Pistoke':
    break
else:
  INSTALLED_APPS.append('pistoke.Pistoke')

if 'pistoke.ohjain.WebsocketOhjain' not in MIDDLEWARE:
  MIDDLEWARE.append('pistoke.ohjain.WebsocketOhjain')



# -*- coding: utf-8 -*-

import asyncio
from contextlib import asynccontextmanager
import functools

from asgiref.sync import markcoroutinefunction

from .tyokalut import Koriste


class Kattelyvirhe(asyncio.CancelledError):
  ''' Websocket-kättely epäonnistui. '''


class YhteysKatkaistiin(asyncio.CancelledError):
  ''' Yhteys katkaistiin asiakaspäästä (websocket.disconnect). '''


class _WebsocketKoriste(Koriste):

  def __new__(cls, websocket, **kwargs):
    # pylint: disable=signature-differs
    _websocket = websocket
    while _websocket is not None:
      if isinstance(_websocket, __class__):
        raise ValueError(
          f'Useita sisäkkäisiä Websocket-protokollamäärityksiä:'
          f' {cls}({type(_websocket)}(...))'
        )
      _websocket = getattr(
        _websocket,
        '__wrapped__',
        None
      )
      # while _websocket is not None
    # Merkitään tulosfunktio asynkroniseksi.
    return markcoroutinefunction(
      super().__new__(cls, websocket, **kwargs)
    )
    # def __new__

  # class _WebsocketKoriste


class _WebsocketYhteys:

  saapuva_kattely = {'type': 'websocket.connect'}
  lahteva_kattely = {'type': 'websocket.accept'}
  saapuva_katkaisu = {'type': 'websocket.disconnect'}
  lahteva_katkaisu = {'type': 'websocket.close'}

  async def _avaa_yhteys(self, request):
    # pylint: disable=protected-access
    saapuva_kattely = await request.receive()
    if saapuva_kattely != self.saapuva_kattely:
      request._katkaistu_vastapaasta.set()
      raise Kattelyvirhe(
        'Avaava kättely epäonnistui: %r' % saapuva_kattely
      )
    await request.send(self.lahteva_kattely)
    # async def _avaa_yhteys

  async def _sulje_yhteys(self, request):
    if not request._katkaistu_vastapaasta.is_set():
      request._katkaistu_tasta_paasta.set()
      await request.send(self.lahteva_katkaisu)
      try:
        saapuva_katkaisu = await asyncio.wait_for(
          request.receive(),
          timeout=0.01
        )
        if saapuva_katkaisu != self.saapuva_katkaisu:
          raise Kattelyvirhe(
            'Sulkeva kättely epäonnistui: %r' % saapuva_katkaisu
          )
      except (asyncio.CancelledError, asyncio.TimeoutError):
        pass
    # async def _sulje_yhteys

  @asynccontextmanager
  async def __call__(
    self, request, *args, **kwargs
  ):
    request._katkaistu_vastapaasta = asyncio.Event()
    request._katkaistu_tasta_paasta = asyncio.Event()

    # pylint: disable=invalid-name
    try:
      await asyncio.shield(self._avaa_yhteys(request))

    except Kattelyvirhe:
      await asyncio.shield(self._sulje_yhteys(request))
      raise

    except asyncio.CancelledError:
      asyncio.tasks.current_task().cancel()

    try:
      yield request

    finally:
      try:
        await asyncio.shield(self._sulje_yhteys(request))
      except asyncio.CancelledError:
        pass
    # async def __call__

  # class _WebsocketYhteys


class _WebsocketProtokolla(_WebsocketYhteys):

  saapuva_sanoma = {'type': 'websocket.receive'}
  lahteva_sanoma = {'type': 'websocket.send'}

  class SyotettaEiLuettu(Exception):
    ''' Näkymä ei lukenut kaikkea sille annettua syötettä. '''

  @asynccontextmanager
  async def __call__(
    self, request, *args, **kwargs
  ):
    # pylint: disable=invalid-name
    async with super().__call__(request):

      syote = asyncio.Queue()

      @functools.wraps(request.receive)
      async def _receive():
        while not request._katkaistu_tasta_paasta.is_set() \
        and not request._katkaistu_vastapaasta.is_set():
          sanoma = await _receive.__wrapped__()
          if sanoma['type'] == self.saapuva_sanoma['type']:
            await syote.put(
              sanoma.get('text', sanoma.get('bytes', None))
            )
          elif sanoma['type'] == self.saapuva_katkaisu['type']:
            request._katkaistu_vastapaasta.set()
            break
          else:
            raise TypeError(repr(sanoma))
        # async def _receive

      @functools.wraps(request.receive)
      async def receive():
        data = await syote.get()
        syote.task_done()
        return data
        # async def receive

      @functools.wraps(request.send)
      async def send(data):
        '''
        Lähetetään annettu data joko tekstinä tai tavujonona.
        '''
        if isinstance(data, str):
          data = {**self.lahteva_sanoma, 'text': data}
        elif isinstance(data, bytearray):
          data = {**self.lahteva_sanoma, 'bytes': bytes(data)}
        elif isinstance(data, bytes):
          data = {**self.lahteva_sanoma, 'bytes': data}
        else:
          raise TypeError(repr(data))
        return await send.__wrapped__(data)
        # async def send

      request.receive = receive
      request.send = send

      try:
        yield request, _receive

      except (YhteysKatkaistiin, asyncio.CancelledError):
        pass

      finally:
        request.receive = receive.__wrapped__
        request.send = send.__wrapped__

        # Varmistetaan, että näkymä luki kaiken syötteen.
        if not syote.empty() and self.SyotettaEiLuettu is not None:
          _s = []
          while not syote.empty():
            _s.append(await syote.get())
          raise self.SyotettaEiLuettu(_s)
        # finally

      # async with super

    # async def __call__

  # class _WebsocketProtokolla


class WebsocketProtokolla(_WebsocketProtokolla, _WebsocketKoriste):
  '''
  Sallitaan vain yksi protokolla per metodi.
  '''
  async def __call__(
    self, request, *args, **kwargs
  ):
    # pylint: disable=invalid-name
    if request.method != 'Websocket':
      # pylint: disable=no-member
      return await self.__wrapped__(
        request, *args, **kwargs
      )

    async with super().__call__(
      request, *args, **kwargs
    ) as (request, _receive):
      kaaritty = asyncio.create_task(
        self.__wrapped__(request, *args, **kwargs)
      )
      receive = asyncio.create_task(_receive())

      @receive.add_done_callback
      def vastaanotto_valmis(__):
        kaaritty.cancel()

      try:
        await kaaritty
      finally:
        try:
          await asyncio.wait_for(receive, timeout=0.01)
        except (asyncio.CancelledError, asyncio.TimeoutError):
          pass
        # finally
      # async with super.__call__

    # async def __call__

  # class WebsocketProtokolla


class WebsocketAliprotokolla(WebsocketProtokolla):

  protokolla = []

  def __new__(cls, *args, **kwargs):
    if not args or not callable(args[0]):
      def wsp(websocket):
        return cls(websocket, *args, **kwargs)
      return wsp
    return super().__new__(cls, args[0])
    # def __new__

  def __init__(self, websocket, *protokolla, **kwargs):
    super().__init__(websocket, **kwargs)
    self.protokolla = protokolla
    # def __init__

  async def _avaa_yhteys(self, request):
    saapuva_kattely = await request.receive()
    if saapuva_kattely != self.saapuva_kattely:
      request._katkaistu_vastapaasta.set()
      request.protokolla = None
      raise Kattelyvirhe(
        'Avaava kättely epäonnistui: %r' % saapuva_kattely
      )

    pyydetty_protokolla = request.scope.get(
      'subprotocols', []
    )
    if self.protokolla or pyydetty_protokolla:
      # pylint: disable=protected-access, no-member
      # pylint: disable=undefined-loop-variable
      for hyvaksytty_protokolla in pyydetty_protokolla:
        if hyvaksytty_protokolla in self.protokolla:
          break
      else:
        # Yhtään yhteensopivaa protokollaa ei löytynyt (tai pyynnöllä
        # ei ollut annettu yhtään protokollaa).
        # Hylätään yhteyspyyntö.
        request.protokolla = None
        raise Kattelyvirhe(
          'Ei yhteensopivaa protokollaa: %r vs. %r' % (
            self.protokolla, pyydetty_protokolla
          )
        )
      # Hyväksytään WS-yhteyspyyntö valittua protokollaa käyttäen.
      await request.send({
        **self.lahteva_kattely,
        'subprotocol': hyvaksytty_protokolla,
      })
      request.protokolla = hyvaksytty_protokolla

    else:
      # Näkymä ei määrittele protokollaa; hyväksytään pyyntö.
      await request.send(self.lahteva_kattely)
      request.protokolla = None
    # async def _avaa_yhteys

  # class WebsocketAliprotokolla



# -*- coding: utf-8 -*-

import asyncio
from contextlib import asynccontextmanager
from dataclasses import dataclass
from urllib.parse import urlparse

from asgiref.sync import async_to_sync, sync_to_async

from django.core.signals import (
  request_finished, request_started,
)
from django.db import close_old_connections
from django.test.client import AsyncClient
from django.test.testcases import SimpleTestCase
from django.utils.functional import classproperty

from pistoke.kasittelija import WebsocketKasittelija
from pistoke.protokolla import _WebsocketProtokolla
from pistoke.pyynto import WebsocketPyynto


class WebsocketPoikkeus:

  class NakymaPaattyi(Exception):
    ''' Näkymä päättyi ennen kuin pääteyhteys alkoi. '''

  class KattelyEpaonnistui(Exception):
    ''' Websocket-kättely epäonnistui. '''

  class Http403(Exception):
    ''' Websocket-yhteyspyyntö epäonnistui. '''

  class PaateyhteysAikakatkaistiin(Exception):
    ''' Websocket-pääteyhteys aikakatkaistiin. '''

  class SyotettaEiLuettu(RuntimeError):
    ''' Websocket-syötettä jäi lukematta näkymän päätyttyä. '''

  class TulostettaEiLuettu(RuntimeError):
    ''' Websocket-tulostetta jäi lukematta pääteyhteyden päätyttyä. '''

  # class WebsocketPoikkeus


class WebsocketPaateKasittelija(WebsocketKasittelija):
  ''' Vrt. AsyncClientHandler '''

  nosta_syotetta_ei_luettu = True

  def __init__(
    self,
    *args,
    enforce_csrf_checks=True,
    **kwargs
  ):
    super().__init__(*args, **kwargs)
    self.enforce_csrf_checks = enforce_csrf_checks
    # def __init__

  async def __call__(self, scope, receive, send):
    request_started.disconnect(close_old_connections)
    try:
      await super().__call__(scope, receive, send)
    finally:
      request_started.connect(close_old_connections)
    # async def __call__

  async def get_response_async(self, request):
    # pylint: disable=protected-access
    request._dont_enforce_csrf_checks = not self.enforce_csrf_checks
    return await super().get_response_async(request)
    # async def get_response_async

  # class WebsocketPaateKasittelija


class WebsocketPaateprotokolla(_WebsocketProtokolla):
  '''
  Käänteinen Websocket-protokolla, so. selaimen / ASGI-palvelimen näkökulma.

  Vrt. `pistoke.protokolla.WebsocketProtokolla`.
  '''
  # Websocket-sanomatyypit ovat päinvastaiset.
  saapuva_kattely = {'type': 'websocket.accept'}
  lahteva_kattely = {'type': 'websocket.connect'}
  saapuva_katkaisu = {'type': 'websocket.close'}
  lahteva_katkaisu = {'type': 'websocket.disconnect'}
  lahteva_sanoma = {'type': 'websocket.receive'}
  saapuva_sanoma = {'type': 'websocket.send'}

  # Nostetaan erillinen poikkeus, kun pääteyhteys ei lue omaa syötettään,
  # so. näkymän tuottamaa tulostetta.
  SyotettaEiLuettu = WebsocketPoikkeus.TulostettaEiLuettu

  async def _avaa_yhteys(self, request):
    ''' Kättelyiden järjestys on päinvastainen. '''
    await request.send(self.lahteva_kattely)
    kattely = await request.receive()
    if not isinstance(kattely, dict) or 'type' not in kattely:
      raise WebsocketPoikkeus.KattelyEpaonnistui(
        'Virheellinen kättely: %r' % kattely
      )
    if kattely == self.saapuva_katkaisu:
      request._katkaistu_vastapaasta.set()
      raise WebsocketPoikkeus.Http403(
        'Palvelin sulki yhteyden.'
      )
    elif kattely['type'] == self.saapuva_kattely['type']:
      if 'subprotocol' in kattely:
        request.scope['subprotocol'] = kattely['subprotocol']
    else:
      raise WebsocketPoikkeus.KattelyEpaonnistui(
        'Virheellinen kättely: %r' % kattely
      )
    # async def _avaa_yhteys

  async def _sulje_yhteys(self, request):
    '''
    Yhteys suljetaan (disconnect) riippumatta mahdollisesta
    näkymän lähettämästä katkaisusta (close).
    '''
    await request.send(self.lahteva_katkaisu)
    # async def _sulje_yhteys

  @asynccontextmanager
  async def __call__(self, scope, receive, send):
    async with super().__call__(
      WebsocketPyynto(scope, receive, send),
    ) as (request, _receive):
      _task = asyncio.tasks.current_task()
      _receive = asyncio.create_task(_receive())
      _receive.add_done_callback(
        lambda __receive: _task.cancel()
      )
      try:
        yield request
      finally:
        _receive.cancel()
        try:
          await _receive
        except asyncio.CancelledError:
          pass
      # async with super().__call__
    # async def __call__

  # class WebsocketPaateprotokolla


@dataclass
class WebsocketPaateyhteys(WebsocketPaateprotokolla):

  scope: dict
  enforce_csrf_checks: bool
  raise_request_exception: bool
  aikakatkaisu: float

  def __post_init__(self):
    super().__init__()

  @asynccontextmanager
  async def __call__(self):
    kasittelija = WebsocketPaateKasittelija(
      enforce_csrf_checks=self.enforce_csrf_checks
    )
    syote, tuloste = asyncio.Queue(), asyncio.Queue()

    nakyma = asyncio.create_task(
      kasittelija(
        self.scope,
        syote.get,
        tuloste.put,
      )
    )

    async def aikakatkaistu_nakyma():
      valmis, kesken = await asyncio.wait(
        (nakyma, ),
        timeout=self.aikakatkaisu,
      )
      if valmis:
        await nakyma
      else:  # if kesken
        nakyma.cancel()
        try:
          await nakyma
        except asyncio.CancelledError:
          pass
        raise WebsocketPoikkeus.PaateyhteysAikakatkaistiin
      # async def aikakatkaistu_nakyma
    aikakatkaistu_nakyma = asyncio.create_task(aikakatkaistu_nakyma())

    paate = asyncio.tasks.current_task()
    paatteen_nostama_poikkeus = None

    @aikakatkaistu_nakyma.add_done_callback
    def nakyma_valmis(_nakyma):
      ''' Keskeytä pääteistunto, jos näkymä päättyy. '''
      paate.cancel()

    # Anna näkymälle hetki aikaa ennen pääteyhteyden avaamista.
    # Mikäli näkymä ehti päättyä, nostetaan poikkeus.
    try:
      await asyncio.sleep(0.01)
    except asyncio.CancelledError:
      try:
        if poikkeus := aikakatkaistu_nakyma.exception():
          raise poikkeus from None
      except asyncio.CancelledError:
        pass
      raise WebsocketPoikkeus.NakymaPaattyi from None

    try:
      # Toteutetaan pääteistunto käänteisen Websocket-protokollan
      # sisällä.
      async with super().__call__(
        self.scope,
        tuloste.get,
        syote.put
      ) as request:
        yield request

    except Exception as exc:
      paatteen_nostama_poikkeus = exc

    finally:
      # Annetaan näkymälle hetki aikaa päättyä.
      try:
        await asyncio.sleep(0.01)
      except asyncio.CancelledError:
        pass

      aikakatkaistu_nakyma.cancel()

      # Odotetaan, kunnes näkymän suoritus on päättynyt.
      try:
        await aikakatkaistu_nakyma
      except asyncio.CancelledError:
        pass
      except _WebsocketProtokolla.SyotettaEiLuettu as exc:
        raise WebsocketPoikkeus.SyotettaEiLuettu from exc

      finally:
        if paatteen_nostama_poikkeus is not None:
          raise paatteen_nostama_poikkeus

      # async with super.__call__
    # async def __call__

  # class WebsocketPaateyhteys


def websocket_scope(
  paate,
  path,
  secure=False,
  protokolla=None,
  **extra
):
  '''
  Muodosta Websocket-pyyntökonteksti (scope).

  Vrt. `django.test.client:AsyncRequestFactory`:
  metodit `_base_scope` ja `generic`.
  '''
  # pylint: disable=protected-access
  parsed = urlparse(str(path))  # path can be lazy.
  request = {
    'path': paate._get_path(parsed),
    'server': ('127.0.0.1', '443' if secure else '80'),
    'scheme': 'wss' if secure else 'ws',
    'headers': [(b'host', b'testserver')],
  }
  request['headers'] += [
    (key.lower().encode('ascii'), value.encode('latin1'))
    for key, value in extra.items()
  ]
  if not request.get('query_string'):
    request['query_string'] = parsed[4]
  if protokolla is not None:
    request['subprotocols'] = (
      [protokolla] if isinstance(protokolla, str)
      else list(protokolla)
    )
  return {
    'type': 'websocket',
    'asgi': {'version': '3.0', 'spec_version': '2.1'},
    'scheme': 'ws',
    'server': ('testserver', 80),
    'client': ('127.0.0.1', 0),
    'headers': [
      (b'sec-websocket-version', b'13'),
      (b'connection', b'keep-alive, Upgrade'),
      *paate.defaults.pop('headers', ()),
      *request.pop('headers', ()),
      (b'cookie', b'; '.join(sorted(
        ('%s=%s' % (morsel.key, morsel.coded_value)).encode('ascii')
        for morsel in paate.cookies.values()
      ))),
      (b'upgrade', b'websocket')
    ],
    **paate.defaults,
    **request,
  }
  # def websocket_scope


class WebsocketPaate(WebsocketPoikkeus, AsyncClient):

  websocket_aikakatkaisu: float = 1.0

  def __init__(self, *args, **kwargs):
    try:
      self.websocket_aikakatkaisu = kwargs.pop('websocket_aikakatkaisu')
    except KeyError:
      pass
    super().__init__(*args, **kwargs)
    # def __init__

  def websocket(self, *args, **kwargs):
    '''
    Käyttö asynkronisena kontekstina:
    >>> class Testi(WebsocketTesti):
    >>>
    >>>   async def testaa_X(self):
    >>>     async with self.async_client.websocket(
    >>>       '/.../'
    >>>     ) as websocket:
    >>>       websocket.send(...)
    >>>       ... = await websocket.receive()

    Annettu testirutiini suoritetaan ympäröivässä kontekstissa
    ja testattava näkymä tausta-ajona (asyncio.Task).
    '''
    # pylint: disable=protected-access

    return WebsocketPaateyhteys(
      websocket_scope(
        self,
        *args,
        **kwargs
      ),
      enforce_csrf_checks=self.handler.enforce_csrf_checks,
      raise_request_exception=self.raise_request_exception,
      aikakatkaisu=self.websocket_aikakatkaisu,
    )()
    # async def websocket

  # class WebsocketPaate


class WebsocketTesti(SimpleTestCase):

  websocket_aikakatkaisu: float = 1.0

  @classproperty
  def async_client_class(cls):
    # pylint: disable=no-self-argument
    class _WebsocketPaate(WebsocketPaate, super().async_client_class):
      websocket_aikakatkaisu = cls.websocket_aikakatkaisu
    return _WebsocketPaate
    # def async_client_class

  # class WebsocketTesti



# -*- coding: utf-8 -*-

from django.utils.functional import classproperty
from django.views import generic


class WebsocketNakyma(generic.View):
  '''
  Saateluokka; suora yksilöinti on kielletty.

  Lisää kunkin periytetyn näkymäluokan `http_method_names`-luetteloon
  tyyppi "websocket".

  Ohita (aina asynkroninen) `websocket`-metodi HTTP-metodien mahdollista
  asynkronisuutta tarkasteltaessa (Django 4.1+).

  Aseta kunkin periytetyn näkymäluokan `dispatch`-metodille määre
  `_websocket_protokolla` silloin, kun luokan `websocket`-metodi
  määrittelee käyttämänsä protokollan.
  '''

  @classmethod
  def __init_subclass__(cls, *args, **kwargs):
    super().__init_subclass__(*args, **kwargs)

    if 'websocket' not in cls.http_method_names:
      cls.http_method_names.append('websocket')

    if hasattr(cls, 'dispatch') \
    and hasattr(cls.websocket, 'protokolla'):
      # pylint: disable=protected-access, no-member
      cls.dispatch._websocket_protokolla = cls.websocket.protokolla
    # def __init_subclass__

  @classproperty
  def view_is_async(cls):
    # pylint: disable=no-self-argument
    http_method_names, cls.http_method_names = cls.http_method_names, [
      http for http in cls.http_method_names
      if http != 'websocket'
    ]
    paluu = super().view_is_async
    cls.http_method_names = http_method_names
    return paluu
    # def view_is_async

  def __new__(cls, *args, **kwargs):
    if cls == __class__:
      raise NotImplementedError
    return super().__new__(cls)
    # def __new__

  async def websocket(self, request, *args, **kwargs):
    raise NotImplementedError

  # class WebsocketNakyma



# -*- coding: utf-8 -*-

import warnings


def json_viestiliikenne(*args, **kwargs):
  ''' Lähetä ja vastaanota JSON-muodossa. '''
  from pistoke.tyokalut import JsonLiikenne
  warnings.warn(
    'pistoke.tyokalut.json_viestiliikenne() on poistunut käytöstä.'
    ' Korvaava koriste: pistoke.tyokalut.JsonLiikenne.',
    DeprecationWarning,
    stacklevel=2,
  )
  return JsonLiikenne(*args, **kwargs)
  # def json_viestiliikenne


def protokolla(*ws_protokolla):
  from pistoke.protokolla import WebsocketAliprotokolla
  warnings.warn(
    'pistoke.tyokalut.protokolla() on poistunut käytöstä.'
    ' Korvaava protokolla: pistoke.protokolla.WebsocketAliprotokolla.',
    DeprecationWarning,
    stacklevel=2,
  )
  return WebsocketAliprotokolla(*ws_protokolla)
  # def protokolla


def csrf_tarkistus(*args, **kwargs):
  '''
  Tarkista ensimmäisen sanoman mukana toimitettu CSRF-tunniste.

  Jos parametri `csrf_avain` on annettu, poimitaan sanakirjamuotoisesta
  syötteestä.
  '''
  from pistoke.tyokalut import CsrfKattely
  warnings.warn(
    'pistoke.tyokalut.csrf_tarkistus() on poistunut käytöstä.'
    ' Korvaava koriste: pistoke.tyokalut.CsrfKattely.',
    DeprecationWarning,
    stacklevel=2,
  )
  return CsrfKattely(*args, **kwargs)
  # def csrf_tarkistus



# -*- coding: utf-8 -*-

# pylint: disable=invalid-name
# pylint: disable=ungrouped-imports

from urllib.parse import urlparse

from django.conf import settings
from django.core.asgi import get_asgi_application
from django.core.handlers.asgi import ASGIHandler
from django.core.management import CommandError

try:
  import uvicorn
except ImportError:
  uvicorn = None

from pistoke.kasittelija import WebsocketKasittelija


if 'django.contrib.staticfiles' in settings.INSTALLED_APPS:
  # Mikäli `staticfiles` on käytössä,
  # (1) periytetään `runserver`-komento
  #     kyseisessä sovelluksessa periytetystä `runserver`-komennosta ja
  # (2) käytetään erillistä ASGI-käsittelijää
  #     staattisten tiedostojen tarjoilemiseen.
  from django.contrib.staticfiles.management.commands.runserver import (
    Command as _Command
  )
  from django.contrib.staticfiles.handlers import StaticFilesHandlerMixin
  from django.utils.cache import add_never_cache_headers
  class static_handler(StaticFilesHandlerMixin, ASGIHandler):
    ''' Vrt. staticfiles.handlers.ASGIStaticFilesHandler '''
    def __init__(self):
      # pylint: disable=super-init-not-called
      # Ei kutsuta `ASGIHandler.__init__`-metodia,
      # joka lataa välikkeet (middleware); niitä ei tarvita.
      self.base_url = urlparse(self.get_base_url())
      # def __init__
    def serve(self, request):
      # Kielletään staattisten tiedostojen paikallinen välimuisti.
      tulos = super().serve(request)
      add_never_cache_headers(tulos)
      return tulos
      # def serve
    # class static_handler
  # if 'django.contrib.staticfiles' in settings.INSTALLED_APPS

else:
  # Muuten periytetään suoraan django.core-versiosta.
  from django.core.management.commands.runserver import (
    Command as _Command
  )
  static_handler = None
  # else


class Kasittelija:
  ''' Käsittele tulevat ASGI-pyynnöt oikean käsittelijän läpi. '''

  def __init__(self, django, websocket, static=False):
    ''' Alusta tyyppikohtaiset käsittelijät. '''
    self.django = django
    self.websocket = websocket
    self.static = static_handler() if static_handler and static else None
    # def __init__

  async def __call__(self, scope, receive, send):
    ''' Välitä pyyntö oikealle käsittelijälle otsaketietojen mukaan. '''
    # pylint: disable=not-callable, protected-access
    if scope['type'] == 'http':
      if self.static and self.static._should_handle(scope['path']):
        return await self.static(scope, receive, send)
      else:
        return await self.django(scope, receive, send)
    elif scope['type'] == 'websocket':
      return await self.websocket(scope, receive, send)
    else:
      raise ValueError(f'tuntematon pyyntö: {scope["type"]}')
    # async def __call__

  # class Kasittelija


# Luodaan kaksi erillistä ASGI-käsittelyfunktiota:
# staticfilesillä ja ilman sitä.
uvicorn_application = Kasittelija(
  django=get_asgi_application(),
  websocket=WebsocketKasittelija(),
)
uvicorn_application_static = Kasittelija(
  django=get_asgi_application(),
  websocket=WebsocketKasittelija(),
  static=True,
)


class Command(_Command):

  def add_arguments(self, parser):
    # pylint: disable=protected-access
    super().add_arguments(parser)
    # Mikäli uvicorn on asennettu, käytetään oletuksena
    # ASGI-palvelinta ja lisätään vipu WSGI:n käyttöön.
    # Lisätään myös uvicorn-lokitustasot mahdollisina
    # komentoriviparametreinä '-v'-vivulle.
    for a in parser._actions:
      if a.dest == 'verbosity':
        a.type = lambda s: int(s) if s in ['0', '1', '2', '3'] else s
        a.choices += [
          'critical', 'error', 'warning', 'info', 'debug', 'trace'
        ]
        break
    # Lisätään parametrit `--wsgi` ja `--asgi` palvelimen tyypin
    # valitsemiseksi.
    oletus_asgi = bool(uvicorn is not None)
    parser.add_argument(
      '--asgi',
      action='store_true',
      dest='asgi',
      default=oletus_asgi,
      help=f'Käynnistä ASGI-palvelin (oletus{"" if oletus_asgi else " WSGI"}).'
    )
    parser.add_argument(
      '--wsgi',
      action='store_false',
      dest='asgi',
      default=oletus_asgi,
      help=f'Käynnistä WSGI-palvelin (oletus{" ASGI" if oletus_asgi else ""}).'
    )
    parser.add_argument(
      '--ssl-keyfile',
      default=getattr(self, 'oletus_ssl_keyfile', None),
      help=f'SSL-avaintiedosto (HTTPS)',
    )
    parser.add_argument(
      '--ssl-certfile',
      default=getattr(self, 'oletus_ssl_certfile', None),
      help=f'SSL-varmennetiedosto (HTTPS)',
    )
    # def add_arguments

  def asgi_run(self, addr, port, asgi_handler, **options):
    if uvicorn is None:
      raise CommandError(
        'ASGI-palvelimen ajaminen edellyttää `uvicorn`-paketin käyttöä.'
        ' Asenna se komennolla `pip install uvicorn`.'
      )
    uvicorn.run(
      asgi_handler,
      host=addr,
      port=port,
      log_level={
        0: 'critical',
        #: 'error',
        #: 'warning',
        1: 'info',
        2: 'debug',
        3: 'trace',
      }.get(options['verbosity'], options['verbosity'] or 'info'),
      reload=options['use_reloader'],
      reload_includes=['*'],
      reload_excludes=['*/.git/*'],
      **(
        {'use_colors': True} if options['force_color']
        else {'use_colors': False} if options['no_color']
        else {}
      ),
      **({
        'ssl_keyfile': options['ssl_keyfile'],
        'ssl_certfile': options['ssl_certfile'],
      } if options['ssl_keyfile'] and options['ssl_certfile'] else {}),
    )
    # def run

  def run(self, **options):
    if options.get('asgi', False) and options['use_reloader']:
      # Ohita mahdollinen automaattinen uudelleenlataus tässä.
      # Käytetään `uvicorn --reload`-parametriä.
      self.inner_run(None, **options)
    else:
      super().run(**options)
    # def run

  class ASGIPalvelin(Exception):
    ''' Poikkeusluokka ASGI-käsittelyn ilmaisemiseksi. '''

  def inner_run(self, *args, **options):
    '''
    Mikäli super() nostaa em. poikkeuksen,
    suoritetaan ASGI-käsittelijä (Uvicorn).
    '''
    try:
      super().inner_run(*args, **options)
    except self.ASGIPalvelin:
      # Poimitaan haluttu sovellus käsillä olevasta moduulista.
      handler = __name__ + (
        ':uvicorn_application_static'
        if options.get('use_static_handler', False) and (
          settings.DEBUG or options['insecure_serving']
        )
        else ':uvicorn_application'
      )
      self.asgi_run(self.addr, int(self.port), handler, **options)
    # def inner_run

  def get_handler(self, *args, **options):
    if options.get('asgi', False):
      # Katkaistaan `super().inner_run`-metodin suoritus ennen
      # WSGI-palvelimen käynnistystä.
      raise self.ASGIPalvelin
    return super().get_handler(*args, **options)
    # def get_handler

  # class Command
